home · contact · privacy
Repaired buggy window title display.
[plomrogue] / windows.c
index b003fd71045ca19eff58c0ea26a9c70c93052e08..31da8333d91c35cd4b3b8671cc37676e99c2a796 100644 (file)
--- a/windows.c
+++ b/windows.c
@@ -6,10 +6,13 @@
 struct WinMeta init_win_meta (WINDOW * screen) {
 // Create and populate WinMeta struct with sane default values.
   struct WinMeta win_meta;
-  win_meta.height = screen->_maxy + 1;
-  win_meta.width = screen->_maxx + 1;
+  win_meta.screen = screen;
+  win_meta.height = getmaxy(screen);
+  win_meta.width = getmaxx(screen);
   win_meta.chain_start = 0;
   win_meta.chain_end = 0;
+  win_meta.pad_offset = 0;
+  win_meta.pad = newpad(win_meta.height, 1);
   return win_meta; }
 
 struct Win init_window (struct WinMeta * win_meta, char * title) {
@@ -53,63 +56,76 @@ void suspend_window (struct WinMeta * win_meta, struct Win * win) {
   win->prev = 0;
   win->next = 0; }
 
-void place_window (struct WinMeta * win_meta, struct Win * win) {
+struct yx place_window (struct WinMeta * win_meta, struct Win * win) {
 // Based on position and sizes of previous window, find fitting place for current window.
-  win->start_x = 0; // if window is first in chain, place it on top-left corner
-  win->start_y = 1;
+  struct yx start;
+  start.x = 0;                                     // if window is first in chain, place it on top-left corner
+  start.y = 1;
   if (0 != win->prev) {
-    if (win->prev->height == win_meta->height - 1)              // if prev window fills entire column, place win
-      win->start_x = win->prev->start_x + win->prev->width + 1; // in new column next to it
+    if (win->prev->height == win_meta->height - 1)                      // if prev window fills entire column,
+      start.x = getbegx(win->prev->curses_win) + win->prev->width + 1;  // place win in new column next to it
     else {
-      struct Win * first_ceiling = win->prev;                        // first_ceiling determines column with;
-      while (first_ceiling->start_y != 1)                            // default: place window in new column
-        first_ceiling = first_ceiling->prev;                         // next to it
-      win->start_x = first_ceiling->start_x + first_ceiling->width + 1;
-      if (first_ceiling->width >= win->width) {   // only place windows in prev column that fit into its width
+      struct Win * first_ceiling = win->prev;                         // first_ceiling determines column with;
+      while (getbegy(first_ceiling->curses_win) != 1)                 // default: place window in new column
+        first_ceiling = first_ceiling->prev;                          // next to it
+      start.x = getbegx(first_ceiling->curses_win) + first_ceiling->width + 1;
+      if (first_ceiling->width >= win->width) {      // only place wins in prev column that fit into its width
         struct Win * win_p = first_ceiling;
         struct Win * lastrow_startwin = win_p;
         while (win_p != win) {
-          if (win_p->start_x == first_ceiling->start_x)
-            lastrow_startwin = win_p;                                   // try to fit window at end of last
-          win_p = win_p ->next; }                                       // row of windows inside column; if
-        int lastcol_start = win->prev->start_x + win->prev->width + 1;  // failure, try opening new row below
-        if (win->width <= first_ceiling->start_x + first_ceiling->width - lastcol_start
+          if (getbegx(win_p->curses_win) == getbegx(first_ceiling->curses_win))
+            lastrow_startwin = win_p;               // try to fit window at the end of the last row of windows
+          win_p = win_p ->next; }                   // inside column; if failure, try opening a new row below
+        int lastcol_start = getbegx(win->prev->curses_win) + win->prev->width + 1;
+        if (win->width <= getbegx(first_ceiling->curses_win) + first_ceiling->width - lastcol_start
             && win->height <= lastrow_startwin->height) {
-          win->start_x = lastcol_start;
-          win->start_y = lastrow_startwin->start_y; }
-        else if (win->height < win_meta->height - (lastrow_startwin->start_y + lastrow_startwin->height)
+          start.x = lastcol_start;
+          start.y = getbegy(lastrow_startwin->curses_win); }
+        else if (win->height < win_meta->height - (getbegy(lastrow_startwin->curses_win) + lastrow_startwin->height)
                  && win->width <= first_ceiling->width) {
-          win->start_x = first_ceiling->start_x;
-          win->start_y = lastrow_startwin->start_y + lastrow_startwin->height + 1; } } } } }
+          start.x = getbegx(first_ceiling->curses_win);
+          start.y = getbegy(lastrow_startwin->curses_win) + lastrow_startwin->height + 1; } } } }
+  return start; }
 
 void update_windows (struct WinMeta * win_meta, struct Win * win) {
// Update geometry of win and its next of kin. Before, destroy window, if visible. After, (re-)build it.
   if (0 != win->curses_win)
     destroy_window (win);
-  place_window(win_meta, win);
-  win->curses_win = newwin(win->height, win->width, win->start_y, win->start_x);
+  struct yx startyx = place_window(win_meta, win);
+  int lastwincol = 0;
+  struct Win * win_p = win_meta->chain_start;
+  while (win_p != 0) {
+    if (win_p != win && getbegx(win_p->curses_win) + win_p->width > lastwincol + 1)
+      lastwincol = getbegx(win_p->curses_win) + win_p->width - 1;
+    else if (win_p == win && startyx.x + win->width > lastwincol + 1)
+      lastwincol = startyx.x + win->width - 1;
+    win_p = win_p->next; }
+  if (getmaxx(win_meta->pad) != lastwincol) {
+    wresize(win_meta->pad, getmaxy(win_meta->pad), lastwincol + 2); }
+  win->curses_win = subpad(win_meta->pad, win->height, win->width, startyx.y, startyx.x);
   if (0 != win->next)
     update_windows (win_meta, win->next); }
 
 void destroy_window (struct Win * win) {
-// Undraw and delete window.
+// Delete window.
   delwin(win->curses_win);
   win->curses_win = 0; }
 
 void draw_window_borders (struct Win * win, char active) {
 // Draw borders of window win, including title. Decorate in a special way if window is marked as active.
   int y, x;
-  for (y = win->start_y; y <= win->start_y + win->height; y++) {
-    mvaddch(y, win->start_x - 1, '|');
-    mvaddch(y, win->start_x + win->width, '|'); }
-  for (x = win->start_x; x <= win->start_x + win->width; x++) {
-    mvaddch(win->start_y - 1, x, '-');
-    mvaddch(win->start_y + win->height, x, '-'); }
+  for (y = getbegy(win->curses_win); y <= getbegy(win->curses_win) + win->height; y++) {
+    mvwaddch(wgetparent(win->curses_win), y, getbegx(win->curses_win) - 1, '|');
+    mvwaddch(wgetparent(win->curses_win), y, getbegx(win->curses_win) + win->width, '|'); }
+  for (x = getbegx(win->curses_win); x <= getbegx(win->curses_win) + win->width; x++) {
+    mvwaddch(wgetparent(win->curses_win), getbegy(win->curses_win) - 1, x, '-');
+    mvwaddch(wgetparent(win->curses_win), getbegy(win->curses_win) + win->height, x, '-'); }
   char min_title_length_visible = 3; // 1 char minimal, plus 2 chars for decoration left/right of title
-  if (win->width > min_title_length_visible) {
-    int title_length = strlen(win->title);
-    int title_offset = (((win->width) - (title_length + 2)) / 2); // + 2 is for decoration
-    int length_visible = strnlen(win->title, win->width - min_title_length_visible);
+  if (win->width >= min_title_length_visible) {
+    int title_offset = 0;
+    if (win->width > strlen(win->title) + 2)
+      title_offset = (win->width - (strlen(win->title) + 2)) / 2; // + 2 is for decoration
+    int length_visible = strnlen(win->title, win->width - 2);
     char title[length_visible + 3];
     char decoration = ' ';
     if (1 == active)
@@ -117,8 +133,7 @@ void draw_window_borders (struct Win * win, char active) {
     memcpy(title + 1, win->title, length_visible);
     title[0] = title[length_visible + 1] = decoration;
     title[length_visible + 2] = '\0';
-    mvaddstr(win->start_y - 1, win->start_x + title_offset, title); }
-  refresh(); }
+    mvwaddstr(wgetparent(win->curses_win), getbegy(win->curses_win) - 1, getbegx(win->curses_win) + title_offset, title); } }
 
 void draw_windows_borders (struct Win * win, struct Win * win_active, struct Corners * corners, int ccount) {
 // Craw draw_window_borders() for all windows in chain from win on. Save current window's border corners.
@@ -126,29 +141,28 @@ void draw_windows_borders (struct Win * win, struct Win * win_active, struct Cor
   if (win == win_active)
     active = 1;
    draw_window_borders(win, active);
-  corners[ccount].tl_y = win->start_y - 1;
-  corners[ccount].tl_x = win->start_x - 1;
-  corners[ccount].tr_y = win->start_y - 1;
-  corners[ccount].tr_x = win->start_x + win->width;
-  corners[ccount].bl_y = win->start_y + win->height;
-  corners[ccount].bl_x = win->start_x - 1;
-  corners[ccount].br_y = win->start_y + win->height;
-  corners[ccount].br_x = win->start_x + win->width;
+  corners[ccount].tl.y = getbegy(win->curses_win) - 1;
+  corners[ccount].tl.x = getbegx(win->curses_win) - 1;
+  corners[ccount].tr.y = getbegy(win->curses_win) - 1;
+  corners[ccount].tr.x = getbegx(win->curses_win) + win->width;
+  corners[ccount].bl.y = getbegy(win->curses_win) + win->height;
+  corners[ccount].bl.x = getbegx(win->curses_win) - 1;
+  corners[ccount].br.y = getbegy(win->curses_win) + win->height;
+  corners[ccount].br.x = getbegx(win->curses_win) + win->width;
   if (0 != win->next) {
     draw_windows_borders (win->next, win_active, corners, ccount + 1); } }
 
 void draw_windows (struct Win * win) {
 // Draw contents of all windows in window chain from win on.
-  draw_window(win);
+  win->draw(win);
   if (0 != win->next) {
     draw_windows (win->next); } }
 
 void draw_all_windows (struct WinMeta * win_meta) {
 // Draw all windows and their borders.
-  int y, x;
-  for (y = 0; y < win_meta->height; y++)
-    for (x = 0; x < win_meta->width; x++)
-    mvaddch(y, x, ' ');
+  erase();
+  wnoutrefresh(win_meta->screen);
+  werase(win_meta->pad);
   if (win_meta->chain_start) {
     int n_wins = 1;
     struct Win * win_p = win_meta->chain_start;
@@ -156,20 +170,17 @@ void draw_all_windows (struct WinMeta * win_meta) {
       win_p = win_p->next;
       n_wins++; }
     struct Corners * all_corners = malloc(sizeof(struct Corners) * n_wins);
-    draw_windows_borders (win_meta->chain_start, win_meta->active, all_corners, 0);
     draw_windows (win_meta->chain_start);
-    for (y = 0; y < n_wins; y++) {
-      mvaddch(all_corners[y].tl_y, all_corners[y].tl_x, 'o');
-      mvaddch(all_corners[y].tr_y, all_corners[y].tr_x, 'o');
-      mvaddch(all_corners[y].bl_y, all_corners[y].bl_x, 'o');
-      mvaddch(all_corners[y].br_y, all_corners[y].br_x, 'o'); }
-    free(all_corners); } }
-
-void draw_window(struct Win * win) {
-// Draw window content if visible.
-  if (win->height > 1 && win->width > 1) ;
-    win->draw(win);
-  wrefresh(win->curses_win); }
+    draw_windows_borders (win_meta->chain_start, win_meta->active, all_corners, 0);
+    int i;
+    for (i = 0; i < n_wins; i++) {
+      mvwaddch(win_meta->pad, all_corners[i].tl.y, all_corners[i].tl.x, '+');
+      mvwaddch(win_meta->pad, all_corners[i].tr.y, all_corners[i].tr.x, '+');
+      mvwaddch(win_meta->pad, all_corners[i].bl.y, all_corners[i].bl.x, '+');
+      mvwaddch(win_meta->pad, all_corners[i].br.y, all_corners[i].br.x, '+'); }
+    pnoutrefresh(win_meta->pad, 0, win_meta->pad_offset, 0, 0, win_meta->height, win_meta->width - 1);
+    free(all_corners); }
+  doupdate(); }
 
 void resize_window (struct WinMeta * win_meta, char change) {
 // Grow or shrink currently active window. Correct its geometry and that of its followers.
@@ -53,63 +56,76 @@ void suspend_window (struct WinMeta * win_meta, struct Win * win) {
 
 void update_windows (struct WinMeta * win_meta, struct Win * win) {
-// Update geometry of win and its next of kin. Before, destroy window, if visible. After, (re-)build it.
+// Update geometry of win and its next of kin. Destroy (if visible), (re-)build window. If need, resize pad.
   if (0 != win->curses_win)
     destroy_window (win);